//
// bstring.h
//
//	String class for Basic to C++ translator
//
// Description:
//
//	This file impliments a string class for my Basic
//	translater using VMS style descriptors and string
//	functions. This allows the programs to interface
//	with the VMS routines as easily as under the
//	real VAX/BASIC compiler.
//
//	This is really nothing more than a wrapper around
//	the STR$xxx functions, and the DSC$DESCRIPTOR_S
//	stuff.
//
// History:
//
//	01.00.00 - 09/19/1996 - Kevin Handy
//		Original version. Very minimal. Assignment and output
//		only exist. All code contained in '.h' file (easier
//		to debug).
//
//	01.02.01 - 09/20/1996 - Kevin Handy
//		Added in '+' operator for BString and char.
//
//	01.02.02 - 01/01/1997 - Kevin Handy
//		Added sever string functions, including left,
//		right, pos, mid, string, space, strlen.
//
//	01.02.03 - 01/04/1997 - Kevin Handy
//		Added comparison operators (Bstring to Bstring)
//
//	01.02.04 - 01/05/1997 - Kevin Handy
//		Added more functions. Qnum, Qnum1
//
//	01.02.05 - 01/26/1997 - Kevin Handy
//		Code for gosub.
//
//	01.03.01 - 01/29/1997 - Kevin Handy
//		More functions! Ascii, Edit, Ert.
//
//	01.03.02 - 02/07/1997 - Kevin Handy
//		Change 'short' to 'unsigned short' for various length
//		variable definitions.
//

#ifndef _bstring_h_
#define _bstring_h_

//
// Include files
//
extern "C"
{
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <setjmp.h>
#include "descrip.h"
#include "strdef.h"
}
#include <iostream.h>

/*
 * Define some simple types that vary from machine to machine
 */
typedef short int intx;		// Default integer type
typedef double floatx;		// Default float type

//
// Hard coded constants
//
static const char BEL = 7;		// Bell
static const char BS = 8;		// Backspace
static const char HT = 9;		// Horizontal tab
static const char LF = 10;		// Line feed
static const char VT = 11;		// Vertical tab
static const char FF = 12;		// Form feed
static const char CR = 13;		// Carriage return
static const char SO = 14;		// Shift out
static const char SI = 15;		// Shift in
static const char ESC = 27;		// Escape
static const char SP = 32;		// Space
static const char DEL = 127;		// Delete
//static const double PI = 3.1415926535;// Pi (Someone else already defined this)

//
// Error routines
//
extern long ErrNum;			// Error Number
extern long ErrLine;			// Line Number

//
// Horrible gosub/return code, but it does work (Icky icky icky)
//
#define BStack(x) jmp_buf BGoStack[x]; int BGoCount = 0;
#define BGosub(x) if (setjmp(BGoStack[BGoCount++]) == 0) { goto x; }
#define BReturn longjmp(BGoStack[--BGoCount], 1);

//
// BString class
//
class BString : public dsc$descriptor_s
{
public:
	BString();
	BString(const char* Text);
	BString(char ch);
	~BString();

	//
	// Operators
	//
	BString& operator = (const char* Text);
	BString& operator = (const BString& Source)
		{ str$copy_dx(this, &Source); return *this; }

	//
	// Functions
	//

	//
	// I/O functions (Hack, needs lots of work)
	//
	friend ostream& operator<< (ostream& os, const BString& bs)
		{ if (bs.dsc$a_pointer != NULL) {for (int loop=0; loop<bs.dsc$w_length; loop++)
			os << bs.dsc$a_pointer[loop];} return os; }
	friend istream& operator>> (istream& is, BString& bs)
		{ char inputstr[128]; is >> inputstr; bs = inputstr; }
};


//
// Empty Constructor
//
BString::BString()
{
// cout << "MakeD" << endl;
	dsc$b_dtype = DSC$K_DTYPE_T;		// Text class
	dsc$b_class = DSC$K_CLASS_D;		// Dynamic
	dsc$w_length = 0;			// Length = 0
	dsc$a_pointer = NULL;			// No allocation
}

//
// Create from a text string
//
BString::BString(const char* Text)
{
	unsigned short Length = strlen(Text);

// cout << "MakeS" << endl;
	dsc$b_dtype = DSC$K_DTYPE_T;		// Text class
	dsc$b_class = DSC$K_CLASS_D;		// Static
	dsc$w_length = 0;			// Length
	dsc$a_pointer = NULL;			// Point at origional string

	str$copy_r(this, &Length, Text);
}

//
// Create from a byte constant
//
BString::BString(char ch)
{
	const unsigned short Length = 1;
	char Text[2];

	dsc$b_dtype = DSC$K_DTYPE_T;		// Text class
	dsc$b_class = DSC$K_CLASS_D;		// Static
	dsc$w_length = 0;			// Length
	dsc$a_pointer = NULL;			// Pointer

	Text[0] = ch;
	Text[1] = '\0';
	str$copy_r(this, &Length, &Text);
}


//
// Destructor
//
BString::~BString()
{
// cout << "Killed" << endl;
	//
	// Release memory allocation if this is a string class
	//
	if ((dsc$b_class == DSC$K_CLASS_D) && (dsc$a_pointer != NULL))
	{
		str$free1_dx(this);
	}
}

//
// Assignment
//
BString& BString::operator = (const char* Text)
{
	unsigned short Length = strlen(Text);

	str$copy_r(this, &Length, Text);
	return *this;
}

//
// Left()
//
inline BString left(const BString& Source, long Length)
{
	BString* NewString = new BString;
	str$left(NewString, &Source, &Length);
	return *NewString;
}

//
// Right()
//
inline BString right(const BString& Source, long Length)
{
	BString* NewString = new BString;
	str$right(NewString, &Source, &Length);
	return *NewString;
}

//
// Pos()
//
inline BString Qseg(const BString& Source, long Start, long End)
{
	BString* NewString = new BString;
	str$pos_extr(NewString, &Source, &Start, &End);
	return *NewString;
}

//
// Mid()
//
inline BString mid(const BString& Source, long Start, long Length)
{
	BString* NewString = new BString;
	str$len_extr(NewString, &Source, &Start, &Length);
	return *NewString;
}

//
// string
//
inline BString Qstring(long Count, int Character)
{
	BString* NewString = new BString;
	char Source = Character;

	str$dupl_char(NewString, &Count, &Source);

	return *NewString;
}

//
// space
//
inline BString Qspace(long Count)
{
	BString* NewString = new BString;
	char XSpace = ' ';

	str$dupl_char(NewString, &Count, &XSpace);

	 return *NewString;
}

//
// Qnum(integer)
//
inline BString Qnum(long Number)
{
	char Result[32];
	BString* NewString = new BString;

	sprintf(Result, "%ld ", Number);
	*NewString = Result;

	return *NewString;
}

//
// Qnum(float)
//
inline BString Qnum(double Number)
{
	char Result[32];
	BString* NewString = new BString;

	sprintf(Result, "%f ", Number);
	*NewString = Result;

	return *NewString;
}

//
// Qnum1(integer)
//
inline BString Qnum1(long Number)
{
	char Result[32];
	BString* NewString = new BString;

	sprintf(Result, "%ld", Number);
	*NewString = Result;

	return *NewString;
}

//
// Qnum1(float)
//
inline BString Qnum1(double Number)
{
	char Result[64];
	BString* NewString = new BString;

	sprintf(Result, "%f", Number);
	*NewString = Result;

	return *NewString;
}

//
// str
//
inline BString str(long Number) { return Qnum1(Number); }
inline BString str(double Number) { return Qnum1(Number); }

//
// Trm
//
inline BString trm(const BString& Source)
{
	BString* NewString = new BString;
	unsigned short Length;

	str$trim(NewString, &Source, &Length);
	return *NewString;
}


//
// strlen
//
inline long strlen(const BString& Source)
{
	unsigned short Length;
	char* Address;

	str$analyze_sdesc(&Source, &Length, &Address);
	return Length;
}


//
// Ascii
//
inline int ascii(const BString& Source)
{
	unsigned short Length;
	char* Address;

	str$analyze_sdesc(&Source, &Length, &Address);
	if ((Address == NULL) || (Length == 0))
	{
		return 0;
	}
	else
	{
		return *Address;
	}
}

//
// Addition
//
inline BString& operator + (const BString& Text1, const BString& Text2)
{
	//
	// Allocate a new string
	//
	BString* NewString = new BString;

	//
	// Combine strings together.
	//	Using copy,aappend instead of concat, because varargs
	//	won't give me information about number of paramaters.
	//
	str$copy_dx(NewString, &Text1);
	str$append(NewString, &Text2);

	//
	// Give back the new string
	//
	return *NewString;
}

inline BString& operator + (const BString& Text1, const char* Text2)
{
	//
	// Allocate a new string
	//
	BString* NewString = new BString;
	BString Transform(Text2);

	//
	// Combine strings together
	//
	str$copy_dx(NewString, &Text1);
	str$append(NewString, &Transform);

	//
	// Give back the string
	//
	return *NewString;
}

inline int operator == (const BString& Text1, const BString& Text2)
{
	return !str$compare_eql(&Text1, &Text2);
}

inline int operator < (const BString& Text1, const BString& Text2)
{
	 return str$compare(&Text1, &Text2) == -1;
}

inline int operator <= (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) <= 0;
}

inline int operator > (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) == 1;
}

inline int operator >= (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) >= 0;
}

inline int operator != (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) != 0;
}

//
// Various string type functions
//
static inline int Ascii(const char x) { return x; }
static inline int Ascii(const char* x) { return *x; }
static inline int Ascii(const BString& Source)
{
	unsigned short Length;
	char* Address;
	str$analyze_sdesc(&Source, &Length, &Address);
	if (Length == 0)
	{
		return 0;
	}
	else
	{
		return *Address;
	}
}

//
// rad()
//
inline BString& Qrad(const unsigned int Number)
{
	const char RadChar[] = " ABCDEFGHIJKLMNOPQRSTUVWXYZ$.?0123456789";
	char Result[4];
	BString* NewString = new BString;

	Result[0] = RadChar[(Number / (40 * 40)) % 40];	// 1st Character
	Result[1] = RadChar[(Number / 40) % 40];	// 2nd Character
	Result[2] = RadChar[Number % 40];		// 3rd character
	Result[3] = '\0';

	*NewString = Result;
	return *NewString;
}

//
// Ert
//	This is a really lousy stub for ert$
//
inline BString& ert(const long Number)
{
	BString* NewString = new BString("Error Message #000000");
	return *NewString;
}

//
// Instr
//
inline int instr(long Start, const BString& BaseString,
	const BString& SearchString)
{
	return str$position(&BaseString, &SearchString, &Start);
}
 
//
// Edit
//	Also used by cvt$$ code.
//
BString& edit(const BString& Source, const int Code)
{
	BString* NewString = new BString;
	unsigned short Length;
	char* Address;
	int ResultPtr = 0;
	int LeadSpace = 0;
	int MultiSpace = 0;
	int QuoteMark = 0;

	str$analyze_sdesc(&Source, &Length, &Address);
	char* Result = new char[Length + 1];

	while (Length--)
	{
		unsigned char OneCh = *Address++;

		//
		// 1 - Discard parity bit
		//
		if (Code & 1)
		{
			OneCh &= 0x7f;
		}

		//
		// 2 - Discard all spaces and tabs
		//
		if ((Code & 2) && ((OneCh == ' ') || (OneCh == '\t')))
		{
			if (QuoteMark == 0)
			{
				continue;
			}
		}

		//
		// 4 - Discard all carriage returns, line feeds,
		// form feeds, deletes, escapes, and nulls
		//
		if ((Code & 4) && ((OneCh == '\r') || (OneCh == '\n') ||
			(OneCh == '\f') || (OneCh == 27) || (OneCh == 0)))
		{
			if (QuoteMark == 0)
			{
				continue;
			}
		}

		//
		// 8 - Discard Leading Spaces and Tabs
		//
		if ((Code & 8) && (LeadSpace == 0))
		{
			if ((OneCh == ' ') || (OneCh == '\t'))
			{
				continue;
			}
			LeadSpace = 1;
		}
		else
		{
			LeadSpace = 0;
		}

		//
		// 16 - Convert Multiple Spaces and Tabs to one space
		//
		if (Code & 16)
		{
			if ((OneCh == ' ') || (OneCh == '\t'))
			{
				if (MultiSpace != 0)
				{
					continue;
				}
				MultiSpace = 1;
			}
			else
			{
				MultiSpace = 0;
			}
		}

		//
		// 32 - Convert lower case to upper case
		//
		if (Code & 32)
		{
			if (QuoteMark == 0)
			{
				OneCh = toupper(OneCh);
			}
		}

		//
		// 64 - Convert left bracket to left parentheses, and
		// right bracket to right parentheses
		//
		if ((Code & 64) && (QuoteMark == 0))
		{
			if (OneCh == '[')
			{
				OneCh = '(';
			}
			if (OneCh == ']')
			{
				OneCh = ')';
			}
		}

		//
		// 256 - Suppress all editing for characters within
		// quotation marks. If the string has only one quotation
		// mark, suppresses all editing for the characters
		// following the quotation mark.
		//
		if (Code & 256)
		{
			if (QuoteMark)
			{
				if (OneCh == QuoteMark)
				{
					QuoteMark = 0;
				}
			}
			else
			{
				if ((OneCh == '"') || (OneCh == '\''))
				{
					QuoteMark = OneCh;
				}
			}
		}

		//
		// If we get here, we want to keep the character
		//
		Result[ResultPtr++] = OneCh;
	}

	//
	// 128 - Discard Trailing Spaces and Tabs
	//
	if ((Code & 128) && (QuoteMark == 0))
	{
		while ((ResultPtr > 0) &&
			((Result[ResultPtr - 1] == ' ') ||
			(Result[ResultPtr - 1] == '\t')))
		{
			ResultPtr--;
		}
	}

	//
	// Terminate the string
	//
	Result[ResultPtr] = '\0';

	//
	// Return the final answer
	//
	*NewString = Result;
	return *NewString;
}

double atof(const BString& Source)
{
	unsigned short Length;
	char* Address;
	double result;

	//
	// Determine string info
	//
	str$analyze_sdesc(&Source, &Length, &Address);

	//
	// make a copy we can do stuff to
	//
	if (Length == 0)
	{
		return 0.0;
	}
	char* Work = (char*)malloc(Length + 1);
	assert(Work != NULL);

	memcpy(Work, Address, Length);
	Work[Length] = '\n';

	//
	// Now we can actually do an atof
	//
	result = atof(Work);

	//
	// Release allocated memory
	//
	free(Work);

	//
	// Return result
	//
	return result;
}

#endif

